import configparser
import json

import torch
from torch.utils.data import DataLoader


class Config(object):
    def __init__(self, cfg) -> None:
        super().__init__()
        if isinstance(cfg, configparser.ConfigParser):
            self.content = Config.from_cfg(cfg)
        elif isinstance(cfg, str):
            self.content = Config.from_ini(cfg)
        elif isinstance(cfg, dict):
            self.content = json.dumps(Config.from_dict(cfg), indent=2)

    @staticmethod
    def from_ini(ini):
        config = configparser.ConfigParser()
        config.read_string(ini)
        return Config.from_cfg(config)

    @staticmethod
    def from_cfg(cfg):
        dic = {}
        sections = cfg.sections()
        for section in sections:
            dic_section = {}
            options = cfg.options(section)
            for option in options:
                dic_section[option] = cfg.get(section, option)
            dic[section] = dic_section
        return dic

    @staticmethod
    def from_dict(dic):
        res = {}
        # only loss function name reserved.
        if isinstance(dic, torch.nn.modules.loss._Loss):
            return dic._get_name()
        #
        if isinstance(dic, torch.optim.Optimizer):
            sub = dic.param_groups[0].copy()
            sub.pop('params')
            sub['name'] = dic.__class__.__name__
            return Config.from_dict(sub)
        for k in dic.keys():
            if type(dic[k]) in [int, float, bool, str, list]:
                res[k] = dic[k]
            elif isinstance(dic[k], (torch.optim.Optimizer,
                                     torch.nn.modules.loss._Loss)):
                res[k] = Config.from_dict(dic[k])
        return res

    def __str__(self):
        return self.content



import os

import torch
from matplotlib.figure import Figure

from torchtracer.data import Config, Model


class StoreMan(object):
    """
    Storage module.
    """
    CONFIG_FILENAME = 'config.json'
    IMG_DIR = 'images'
    MODEL_DESCRIPTION_FILENAME = 'model.txt'
    MODEL_PARAMETERS_FILENAME = 'model.pth'
    LOG_FILENAME = 'log'

    def __init__(self, root, task_id) -> None:
        super().__init__()
        self.active_log = None
        self.root = root
        self.regist(task_id)

    def regist(self, task_id) -> None:
        dir_path = os.path.abspath(os.path.join(self.root, task_id))
        if os.path.isdir(dir_path):
            raise FileExistsError('{} exists, you should rename the task id.'.format(task_id))
        self.root = self.mkdir(dir_path)

    def close(self):
        self.active_log.close()

    @staticmethod
    def mkdir(path) -> os.path:
        # create task directory
        if not os.path.isdir(path):
            os.mkdir(path)
        # create image directory
        img_dir = os.path.join(path, StoreMan.IMG_DIR)
        if not os.path.isdir(img_dir):
            os.mkdir(img_dir)
        return path

    def store(self, item, file):
        if isinstance(item, Config):
            self.store_config(item)
        elif isinstance(item, Model):
            self.store_model(item, file)
        elif isinstance(item, Figure):
            self.store_image(item, file)

    def store_config(self, cfg):
        path = os.path.join(self.root, self.CONFIG_FILENAME)
        with open(path, 'w') as f:
            f.write(str(cfg))

    def store_model(self, model, file=None):
        description = str(model)
        parameters = model.state_dict

        description_file = os.path.join(self.root,
                                        '{}.txt'.format(file) if file else self.MODEL_DESCRIPTION_FILENAME)
        parameters_file = os.path.join(self.root,
                                       '{}.pth'.format(file) if file else self.MODEL_PARAMETERS_FILENAME)

        with open(description_file, 'w', encoding='utf-8') as f:
            f.write(description)
        torch.save(parameters, parameters_file)

    def store_image(self, image, file):
        img_file = os.path.join(self.root, self.IMG_DIR, file)
        image.savefig(img_file)

    def log(self, msg, file=None):
        logfile = os.path.join(self.root,
                               '{}.{}'.format(file, self.LOG_FILENAME) if file else self.LOG_FILENAME)
        if self.active_log:
            if not os.path.basename(self.active_log.name) == logfile:
                self.active_log.close()
                self.active_log = open(logfile, 'a', encoding='utf-8')
        else:
            self.active_log = open(logfile, 'a', encoding='utf-8')
        self.active_log.write(msg + '\n')



class Model(object):

    def __init__(self, model) -> None:
        super().__init__()
        # print(model)
        self.state_dict = model.state_dict()
        self.name = model.__class__.__name__
        self.architecture = str(model)

    def __str__(self):
        return '{0.name}\n{0.architecture}'.format(self)



import setuptools


with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='torchtracer',
    version='0.1.5',
    author='OIdiotLin',
    author_email='oidiotlin@gmail.com',
    maintainer='OIdiotLin',
    maintainer_email='oidiotlin@gmail.com',
    description='A python package for visualization and storage management in a pytorch AI task.',
    license='MIT License',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/OIdiotLin/torchtracer',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'matplotlib',
        'numpy',
    ]
)



import os
from datetime import datetime

from torchtracer.utils import StoreMan


class Tracer:
    def __init__(self, root):
        self.root = root
        self.storage = None
        print('Tracer start at {}'.format(os.path.abspath(self.root)))

    def attach(self, task_id=None):
        if task_id is None:
            task_id = datetime.now().isoformat(sep='T', timespec='minutes')
        self.storage = StoreMan(self.root, task_id)
        print('Tracer attached with task: {}'.format(task_id))
        return self

    def detach(self):
        self.storage.close()

    def store(self, item, file=None):
        if self.storage is None:
            raise Exception('You should attach with task id first.')
        self.storage.store(item, file)

    def log(self, msg, file=None):
        if self.storage is None:
            raise Exception('You should attach with task id first.')
        self.storage.log(msg, file)
